#!/usr/bin/python
#
# replacement for sudo's missing -A option on old redhat

import os
import sys
import pty
import tty
from select import select
from time import time

def get_output(shellcmd):
    p = os.popen(shellcmd, 'r')
    output = p.read()
    stat = p.close() or 0
    if not os.WIFEXITED(stat) or os.WEXITSTATUS(stat) != 0:
        raise OSError("failed to run %s: exit code 0x%x" % (shellcmd, stat))
    return output

def process_children(pid):
    try:
        return get_output('ps --ppid %d -o pid= -o args=' % pid).splitlines()
    except OSError:
        return ()

noop = lambda *_:0

# states
WAITING_FOR_PROMPT  = 0
PASSWORD_SENT       = 1
CHILDPROC_RUNNING   = 2
CHILD_EXITED        = 3

class FunctionalAsyncore:
    default_poll_period = 5.0
    blocksize = 4096

    def __init__(self):
        self.listeners = {}
        self.exit_watch = {}
        self.parenthood_watch = {}
        self.delayed_calls = []
        self.write_buffers = {}
        self.remove_when_empty = set()
        self.debug = noop

    def loop(self):
        while True:
            if not self.find_things_to_do():
                break

    def find_things_to_do(self):
        self.check_children()
        self.check_grandchildren()
        self.run_delayed_calls()

        if not self.listeners \
        and not self.write_buffers \
        and not self.delayed_calls \
        and not self.exit_watch \
        and not self.parenthood_watch:
            return False

        timeout = self.default_poll_period
        if self.delayed_calls:
            timeout = max(0, min(timeout, time() - self.delayed_calls[0][0]))

        self.debug('selecting on %r, %r; timeout %r'
                   % (self.listeners.keys(), self.write_buffers.keys(),
                      timeout))
        rfds, wfds, xfds = select(self.listeners.keys(),
                                  self.write_buffers.keys(),
                                  [],
                                  timeout)
        self.debug('available for action: %r, %r' % (rfds, wfds))

        self.check_children()
        self.check_listeners(rfds)
        self.check_writers(wfds)
        return True

    def check_children(self):
        for cpid, cb in self.exit_watch.items():
            pid, childstat = os.waitpid(cpid, os.WNOHANG)
            if pid != 0:
                self.debug('child %d exited' % cpid)
                cb(childstat)
                del self.exit_watch[cpid]
                # also remove from parenthood watch, if it's there
                self.parenthood_watch.pop(cpid, None)

    def check_grandchildren(self):
        for cpid, cb in self.parenthood_watch.items():
            children = process_children(cpid)
            if children:
                cb(children)
                del self.parenthood_watch[cpid]

    def run_delayed_calls(self):
        while self.delayed_calls and self.delayed_calls[0][0] < time():
            self.delayed_calls[0][1]()
            self.delayed_calls = self.delayed_calls[1:]

    def check_listeners(self, rfds):
        for finishing_fd in list(self.remove_when_empty):
            if finishing_fd not in rfds:
                self.remove_when_empty.remove(finishing_fd)
                self.stop_listening(finishing_fd)
        for readme in rfds:
            try:
                cb = self.listeners[readme]
                data = os.read(readme, self.blocksize)
            except KeyError:
                continue
            except OSError, e:
                self.debug('OSError on %r (%s)' % (readme, e))
                data = ''
            cb(data)
            if data == '':
                self.debug('EOF on %r. Removing from readers.' % readme)
                del self.listeners[readme]

    def check_writers(self, wfds):
        for writeme in wfds:
            dats = ''.join(self.write_buffers[writeme])
            n = os.write(writeme, dats)
            dats = dats[n:]
            if dats == '':
                self.debug('Done writing to %r. Removing from writers.' % writeme)
                del self.write_buffers[writeme]
            else:
                self.write_buffers[writeme] = [dats]

    def write(self, fd, data):
        buf = self.write_buffers.setdefault(fd, [])
        if buf:
            self.debug('Appending to write buffer for %d.' % fd)
        else:
            self.debug('Starting to write to %d.' % fd)
        buf.append(data)

    def listen(self, fd, callback):
        old = self.listeners.get(fd)
        self.listeners[fd] = callback
        return old

    def call_later(self, seconds, cb):
        new_delayed = (time() + seconds, cb)
        self.delayed_calls.append(new_delayed)
        self.delayed_calls.sort()
        def remove_call():
            self.delayed_calls = [c for c in self.delayed_calls if c is not new_delayed]
        return remove_call

    def stop_listening(self, fd):
        self.debug('Asked to stop listening on %d' % fd)
        self.listeners.pop(fd, None)

    def stop_listening_when_empty(self, fd):
        self.debug('Asked to stop listening on %d when empty' % fd)
        self.remove_when_empty.add(fd)

    def cancel_writer(self, fd):
        self.write_buffers.pop(fd, None)

    def watch_for_process_children(self, pid, cb):
        self.parenthood_watch[pid] = cb

    def watch_for_child_exit(self, pid, cb):
        self.exit_watch[pid] = cb

    def stop_watching_pid(self, pid):
        self.parenthood_watch.pop(pid, None)
        self.exit_watch.pop(pid, None)

class SudoStateMachine:
    def __init__(self, path_to_sudo, cmd_args, pwd):
        self.path_to_sudo = path_to_sudo
        self.cmd_args = cmd_args
        self.pwd = pwd
        self.debug = noop
        self.async = FunctionalAsyncore()
        if os.environ.get('SUDOWRAP_DEBUG', ''):
            self.debug = self.print_debug
            self.async.debug = self.print_debug

    def print_debug(self, msg):
        sys.stderr.write(msg + '\n')

    def initialize_state(self, pid, master_fd, stdout, stderr):
        self.pid = pid
        self.fd = master_fd
        self.child_stdout = stdout
        self.child_stderr = stderr
        self.tty_output = ''
        self.eat_tty_newline = False
        self.error_output = ''
        self.child_status = None
        self.state = WAITING_FOR_PROMPT
        self.cancel_confirm_pw = noop
        self.cancel_confirm_sudo_mad = noop

    def initialize_watchers(self, a):
        a.listen(self.fd, self.got_tty_output)
        a.listen(self.child_stdout, self.got_stdout_from_child)
        a.listen(self.child_stderr, self.got_stderr_from_child)
        a.watch_for_process_children(self.pid, self.sudo_has_a_child)
        a.watch_for_child_exit(self.pid, self.sudo_died)

    def run(self):
        subproc_stdout = os.pipe()
        subproc_stderr = os.pipe()
        pid, master_fd = pty.fork()

        if pid == 0:
            os.dup2(subproc_stdout[1], 1)
            os.dup2(subproc_stderr[1], 2)
            os.close(subproc_stdout[0])
            os.close(subproc_stderr[0])
            os.execlp(self.path_to_sudo, self.path_to_sudo, *self.cmd_args)
            os._exit(1)

        try:
            os.close(subproc_stdout[1])
            os.close(subproc_stderr[1])
            self.initialize_state(pid, master_fd, subproc_stdout[0], subproc_stderr[0])
            self.initialize_watchers(self.async)
            self.async.loop()
        finally:
            try:
                os.close(master_fd)
            except OSError:
                pass

        if os.WIFEXITED(self.child_status):
            stat = os.WEXITSTATUS(self.child_status)
        elif os.WIFSIGNALED(self.child_status):
            stat = -os.WTERMSIG(self.child_status)
        else:
            stat = -1
        return stat

    def got_tty_output(self, dat):
        self.tty_output += dat
        if self.eat_tty_newline:
            if '\n' in self.tty_output:
                parts = self.tty_output.split('\n', 1)
                self.tty_output = ''
                self.eat_tty_newline = False
                if parts[1]:
                    self.got_tty_output(parts[1])
                return

        if self.state in (CHILDPROC_RUNNING, CHILD_EXITED):
            self.debug('tty output in state %d: %r' % (self.state, dat))
            return self.async.write(2, dat)

        self.cancel_confirm_pw()
        self.cancel_confirm_sudo_mad()

        if self.state == WAITING_FOR_PROMPT:
            if self.tty_output.endswith(': ') or self.tty_output.endswith(':'):
                self.cancel_confirm_pw = self.async.call_later(0.1, self.got_pw_prompt)
        elif self.state == PASSWORD_SENT:
            # still apparently no child under sudo. it might have ended really
            # quickly, or sudo might be upset about a bad password or
            # something. if sudo's child is already gone, then sudo itself
            # will likely exit very soon.
            self.cancel_confirm_sudo_mad = self.async.call_later(0.5, self.sudo_is_mad)

    def got_pw_prompt(self):
        # something wrote to the tty, stopped after ":" or ": ", and hasn't
        # written anything else for a tenth-second, and sudo has no child
        # process. Definitely a password prompt.
        self.async.write(self.fd, self.pwd + '\n')
        self.tty_output = ''
        self.error_output = ''
        self.eat_tty_newline = True
        self.setstate(PASSWORD_SENT)

    def got_stdin(self, dat):
        self.async.write(self.fd, dat)

    def got_stdout_from_child(self, dat):
        # I don't think sudo is ever expected to send output on stdout.
        if self.state != CHILD_EXITED:
            self.setstate(CHILDPROC_RUNNING)
        self.async.write(1, dat)

    def got_stderr_from_child(self, dat):
        self.debug('stderr from child: %r' % dat)
        if self.state in (CHILDPROC_RUNNING, CHILD_EXITED):
            self.async.write(2, dat)
        else:
            self.error_output += dat

    def sudo_is_mad(self):
        # sudo didn't like our password. don't bother trying to send again.
        # can't kill it, cause it's setuid
        map(self.async.stop_listening, (0, self.child_stdout, self.child_stderr, self.fd))
        os.close(self.fd)
        os.close(self.child_stdout)
        os.close(self.child_stderr)
        self.async.stop_watching_pid(self.pid)
        self.child_status = 1
        self.setstate(CHILD_EXITED)  # not really

    def sudo_has_a_child(self, children):
        # sudo has started the child. don't need password anymore.
        # TODO: are there versions of sudo which exec the child instead
        # of forking it? we'd want to check /proc/$pid/cmdline for 'sudo'
        self.debug('sudo process has children (%r)! start child-running mode.'
                   % children)
        self.setstate(CHILDPROC_RUNNING)

    def sudo_died(self, stat):
        self.setstate(CHILD_EXITED)
        self.debug('child exited with status 0x%x' % (stat,))
        self.sudo_pid = None
        self.child_status = stat

    def setstate(self, newstate):
        oldstate = self.state
        self.state = newstate
        if oldstate != newstate:
            self.on_state_change(newstate)

    def on_state_change(self, newstate):
        self.debug('changing to state %r' % newstate)

        if newstate == CHILDPROC_RUNNING:
            self.async.listen(0, self.got_stdin)
        if newstate == CHILD_EXITED:
            # stop trying to read from stdin
            self.async.stop_listening(0)
            # stop trying to write to tty fd
            self.async.cancel_writer(self.fd)
            # stop trying to read from child- but only once there's nothing
            # available there, since things can still be in the pipeline after
            # the child exits
            self.async.stop_listening_when_empty(self.fd)
            self.async.stop_listening_when_empty(self.child_stdout)
            self.async.stop_listening_when_empty(self.child_stderr)
        if newstate in (CHILDPROC_RUNNING, CHILD_EXITED):
            self.cancel_confirm_sudo_mad()
            if self.error_output:
                self.async.write(2, self.error_output)
                self.error_output = ''

if __name__ == "__main__":
    passwd = os.environ.get('SUDO_PASSWORD')
    if passwd is None:
        passwd_prog = os.environ.get('SUDO_ASKPASS')
        if passwd_prog is None:
            passwd = ''
        else:
            try:
                passwd = get_output(passwd_prog).rstrip('\n')
            except OSError, e:
                sys.exit(e)

    args = sys.argv[1:]
    if args[0] == '-A':
        # ignore
        args = args[1:]

    sudo = os.environ.get('SUDO_EXE', 'sudo')
    sm = SudoStateMachine(sudo, args, passwd)
    try:
        sys.exit(sm.run())
    except KeyboardInterrupt:
        sys.exit(-2)
